/* -*- c++ -*- */
/*
 * Copyright 2015 <Arnaud Durand>.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif

#include <gnuradio/io_signature.h>
#include "pfb_channelizer_ccf_impl.h"

void print_device_info(cl_device_id device)
{
  char query_buffer[1024];
  int query_int;
  cl_int err;
  err = clGetDeviceInfo(device, CL_DEVICE_NAME,
    sizeof(query_buffer),
    &query_buffer, NULL);
  printf("CL_DEVICE_NAME: %s\n", query_buffer);
  query_buffer[0] = '\0';
  err = clGetDeviceInfo(device, CL_DEVICE_VENDOR,
    sizeof(query_buffer), &query_buffer,
    NULL);
  printf("CL_DEVICE_VENDOR: %s\n", query_buffer);
  query_buffer[0] = '\0';
  err = clGetDeviceInfo(device, CL_DRIVER_VERSION,
    sizeof(query_buffer), &query_buffer,
    NULL);
  printf("CL_DRIVER_VERSION: %s\n", query_buffer);
  query_buffer[0] = '\0';
  err = clGetDeviceInfo(device, CL_DEVICE_VERSION,
    sizeof(query_buffer), &query_buffer,
    NULL);
  printf("CL_DEVICE_VERSION: %s\n", query_buffer);
  query_buffer[0] = '\0';
  err = clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS,
    sizeof(int), &query_int, NULL);
  printf("CL_DEVICE_MAX_COMPUTE_UNITS: %d\n", query_int);
}

namespace gr {
  namespace filter_opencl {

    void
    pfb_channelizer_ccf_impl::set_taps(const std::vector<float> &taps)
    {
      // Free the taps if already allocated
      /**TODO**/

      unsigned int ntaps = taps.size();
      //d_taps = taps;
      d_taps_per_filter = (unsigned int)ceil((double)ntaps/(double)d_nfilts);
      unsigned int ndtaps = d_nfilts*d_taps_per_filter;
      d_taps = (float*)malloc(sizeof(float) * (d_nfilts*d_taps_per_filter));
      // Make a vector of the taps plus fill it out with 0's to fill
      // each polyphase filter with exactly d_taps_per_filter
      memset(d_taps, 0/*0.0f=0x0000*/, sizeof(float) * (d_nfilts*d_taps_per_filter));

      printf("Partition the filter...\n");

      std::vector<float> tmp_taps = taps;

      //std::reverse(tmp_taps.begin(), tmp_taps.end());
      while((float)(tmp_taps.size()) < d_nfilts*d_taps_per_filter) {
        tmp_taps.push_back(0.0);
      }

      // Partition the filter
      for(int i = 0; i < d_nfilts; i++) {
        for(int j = 0; j < d_taps_per_filter; j++) {
          if(i + j*d_nfilts>d_nfilts*d_taps_per_filter-1)
            continue;
          d_taps[i*d_taps_per_filter+j] = tmp_taps[i + j*d_nfilts];
        }
      }

      std::reverse(d_taps,d_taps+d_nfilts*d_taps_per_filter);
      for(int i = 0; i < d_nfilts; i++){
        float *istart = d_taps+i*d_taps_per_filter, *iend = istart+d_taps_per_filter;
        //std::reverse(istart, iend);
      }

      set_history(d_taps_per_filter);
      printf("d_taps_per_filter: %i\n", d_taps_per_filter);
    }

        void
    pfb_channelizer_ccf_impl::set_channel_map(const std::vector<int> &map)
    {
      gr::thread::scoped_lock guard(d_mutex);

      if(map.size() > 0) {
	unsigned int max = (unsigned int)*std::max_element(map.begin(), map.end());
	if(max >= d_nfilts) {
	  throw std::invalid_argument("pfb_channelizer_ccf_impl::set_channel_map: map range out of bounds.\n");
	}
	d_channel_map = map;
      }
    }

    std::vector<int>
    pfb_channelizer_ccf_impl::channel_map() const
    {
      return d_channel_map;
    }

    pfb_channelizer_ccf::sptr
    pfb_channelizer_ccf::make(unsigned int nfilts, const std::vector<float> &taps)
    {
      return gnuradio::get_initial_sptr
        (new pfb_channelizer_ccf_impl(nfilts, taps));
    }

    /*
     * The private constructor
     */
    pfb_channelizer_ccf_impl::pfb_channelizer_ccf_impl(unsigned int nfilts, const std::vector<float> &taps)
      : gr::block("pfb_channelizer_ccf",
              io_signature::make(nfilts, nfilts, sizeof(gr_complex)),
              io_signature::make(1, nfilts, sizeof(gr_complex)))
    {
      d_nfilts = nfilts;

      if(0/**TODO test if radices are supported**/)
        throw std::invalid_argument("pfb_channelizer: nfilts must be any mix of powers of 2, 3, and 5");

      // Default channel map. The channel map specifies which input
      // goes to which output channel; so out[0] comes from
      // channel_map[0].
      d_channel_map.resize(d_nfilts);
      for(unsigned int i = 0; i < d_nfilts; i++) {
	d_channel_map[i] = i;
      }

      //set_output_multiple(d_nfilts);
      set_taps(taps);

      // Allocate space for vector C
      out_ptr = (gr_complex*)malloc(VECTOR_SIZE*d_nfilts * sizeof(gr_complex));

      // Get platform and device information
      // Set up the Platform
      err = clGetPlatformIDs(0, NULL, &num_platforms);
      platforms = (cl_platform_id *)malloc(sizeof(cl_platform_id)*num_platforms);
      err = clGetPlatformIDs(num_platforms, platforms, NULL);
      //Get the devices list and choose the device you want to run on
      cl_device_id *device_list = NULL;
      err = clGetDeviceIDs( platforms[0], CL_DEVICE_TYPE_CPU, 0, NULL, &num_devices);
      device_list = (cl_device_id *)malloc(sizeof(cl_device_id)*num_devices);
      err = clGetDeviceIDs( platforms[0], CL_DEVICE_TYPE_CPU, num_devices, device_list, NULL);
      /**DEBUG**/
      print_device_info(device_list[0]);
      // Create one OpenCL context for each device in the platform
      context = clCreateContext( NULL, num_devices, device_list, NULL, NULL, &err);

      // Create a command queue
      command_queue = clCreateCommandQueue(context, device_list[0], 0, &err);

      // Create memory buffers on the device for each vector
      in_clmem = clCreateBuffer(context, CL_MEM_READ_ONLY, (VECTOR_SIZE+d_taps_per_filter)*d_nfilts * sizeof(gr_complex), NULL, &err);
      taps_clmem = clCreateBuffer(context, CL_MEM_READ_ONLY, d_nfilts*d_taps_per_filter * sizeof(float), NULL, &err);
      out_clmem = clCreateBuffer(context, CL_MEM_READ_WRITE, VECTOR_SIZE*d_nfilts * sizeof(gr_complex), NULL, &err);

      // Copy the taps to the device
      //err = clEnqueueWriteBuffer(command_queue, in_clmem, CL_TRUE, 0, VECTOR_SIZE * sizeof(float), in_ptr, 0, NULL, NULL);
      err = clEnqueueWriteBuffer(command_queue, taps_clmem, CL_TRUE, 0, d_nfilts*d_taps_per_filter * sizeof(float), d_taps, 0, NULL, NULL);

      /*// Create a program from the kernel source
      FILE *fp;
      // Open the file for reading
      fp = fopen("./lib/fir_kernel.c", "r");
      if (!fp) {
          printf("Error: Failed to read kernel file!\n");
          exit(1);
      }
      fseek(fp,0L,SEEK_END);
      size_t file_size = ftell(fp);
      rewind(fp);
      char * fir_kernel = new char [file_size];
      fread(fir_kernel,file_size,1,fp);
      fir_kernel[file_size] = '\0';*/

      program = clCreateProgramWithSource(context, 1, (const char **)&fir_kernel, NULL, &err);

      // Build the program
      err = clBuildProgram(program, 1, device_list, NULL, NULL, NULL);
      /**DEBUG**/
      if (err == CL_BUILD_PROGRAM_FAILURE) {
        // Determine the size of the log
        size_t log_size;
        clGetProgramBuildInfo(program, device_list[0], CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);

        // Allocate memory for the log
        char *log = (char *) malloc(log_size);

        // Get the log
        clGetProgramBuildInfo(program, device_list[0], CL_PROGRAM_BUILD_LOG, log_size, log, NULL);

        // Print the log
        printf("%s\n", log);
      }

      // Create the OpenCL kernel
      kernel = clCreateKernel(program, "fir_kernel", &err);

      global_size[0] = d_nfilts; // Process the entire lists
      //local_size = 64;

      // Set the arguments of the kernel
      err = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&in_clmem);
      err = clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&taps_clmem);
      err = clSetKernelArg(kernel, 2, sizeof(cl_mem), (void *)&out_clmem);
      // Execute the OpenCL kernel on the list

      // FFT library realted declarations
      clfftDim dim = CLFFT_1D;
      size_t cl_lengths[1] = {d_nfilts};

      // Setup clFFT.
      clfftSetupData fft_setup;
      err = clfftInitSetupData(&fft_setup);
      err = clfftSetup(&fft_setup);

      // Create a default plan for a complex FFT.
      err = clfftCreateDefaultPlan(&plan_handle, context, dim, cl_lengths);

      // Set plan parameters.
      err = clfftSetPlanPrecision(plan_handle, CLFFT_SINGLE);
      //err = clfftSetPlanScale(plan_handle, CLFFT_FORWARD, sqrt(1.0f)/d_nfilts);
      err = clfftSetLayout(plan_handle, CLFFT_COMPLEX_INTERLEAVED, CLFFT_COMPLEX_INTERLEAVED);
      err = clfftSetResultLocation(plan_handle, CLFFT_INPLACE);
      err = clfftSetPlanBatchSize(plan_handle, VECTOR_SIZE);

      /**DEBUG*/
      int query_int;
      err = clGetDeviceInfo(device_list[0], CL_DEVICE_MEM_BASE_ADDR_ALIGN,
      sizeof(int), &query_int, NULL);
      printf("CL_DEVICE_MEM_BASE_ADDR_ALIGN: %d\n", query_int);

      // Bake the plan.
      err = clfftBakePlan(plan_handle, 1, &command_queue, NULL, NULL);

      // Clean up and wait for all the comands to complete
      err = clFlush(command_queue);
      //err = clFinish(command_queue);
    }

    /*
     * Our virtual destructor.
     */
    pfb_channelizer_ccf_impl::~pfb_channelizer_ccf_impl()
    {
      err = clFinish(command_queue);

      // Release the plan.
      err = clfftDestroyPlan( &plan_handle );

      // Release clFFT library.
      clfftTeardown( );

      // Finally release all OpenCL allocated objects and host buffers
      err = clReleaseKernel(kernel);
      err = clReleaseProgram(program);
      err = clReleaseMemObject(in_clmem);
      err = clReleaseMemObject(taps_clmem);
      err = clReleaseMemObject(out_clmem);
      err = clReleaseCommandQueue(command_queue);
      err = clReleaseContext(context);

      free(d_taps);
      free(out_ptr);
      free(platforms);
      free(device_list);
    }

    /*void
    pfb_channelizer_ccf_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    {
        // <+forecast+> e.g. ninput_items_required[0] = noutput_items
    }*/

    int
    pfb_channelizer_ccf_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
      gr::thread::scoped_lock guard(d_mutex);

      gr_complex **inv = (gr_complex**)&input_items[0];
      gr_complex *in = (gr_complex*)input_items[0];
      //gr_complex **outv = (gr_complex **)&output_items[0];
      gr_complex *out = (gr_complex*)output_items[0];

      size_t noutputs = output_items.size();

      for(int i=0; i<d_nfilts; i++)
          err = clEnqueueWriteBuffer(command_queue, in_clmem, CL_FALSE, i * (VECTOR_SIZE+d_taps_per_filter) * sizeof(gr_complex), ninput_items[i] * sizeof(gr_complex), (void *)inv++[0]/**WARN**/, 0, NULL, NULL);

      global_size[1] = noutput_items;
      err = clSetKernelArg(kernel, 3, sizeof(int), (void *)&d_taps_per_filter);

      // Process one item at a time
      err = clEnqueueNDRangeKernel(command_queue, kernel, 2, NULL, global_size, 0/*&local_size*/, 0, NULL, NULL);

      // Execute the plan.
      err = clfftEnqueueTransform(plan_handle, CLFFT_FORWARD, 1, &command_queue, 0, NULL, NULL, &out_clmem, NULL, NULL);

      // Read the cl memory out_clmem on device to the host variable out_ptr
      err = clEnqueueReadBuffer(command_queue, out_clmem, CL_TRUE, 0, d_nfilts*noutput_items * sizeof(gr_complex), out_ptr, 0, NULL, NULL);
      if (err != CL_SUCCESS) {
          printf("Error: Failed to read output array!\n");
          exit(1);
      }

      // Send to output channels
      gr_complex *outbuf=out_ptr;
      for (int oo = 0; oo < noutput_items; oo++) {
        for (unsigned int nn = 0; nn < noutputs/*d_nfilts*/; nn++) {
          out = (gr_complex*)output_items[nn];
          out[oo] = outbuf[d_channel_map[nn]];
        }
        outbuf+=noutputs;
      }

      // Tell runtime system how many input items we consumed on
      // each input stream.
      consume_each (noutput_items);

      // Tell runtime system how many output items we produced.
      return noutput_items;
    }

  } /* namespace filter_opencl */
} /* namespace gr */

